/**************************************************************************
*                                                                         *
* RFDump                                                                  * 
* Copyright (c) 2005 DN-Systems GmbH                                      *
*                                                                         *
***************************************************************************/

#include <stdio.h>

#include <gtk/gtk.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include "interface.h"
#include "support.h"
#include "rfid.h"
#include "settings.h"
#include "tagtypes.h"

gint timeout_callback(gpointer data);
gboolean delete_event(GtkWidget *widget, GdkEvent *event, gpointer data);
void cell_edited_callback(GtkCellRendererText *cell, gchar *path_string, gchar *new_text,gpointer user_data);
int getCharValue(char c);
void updateGUI(int adr, gchar *new_text);

/* POSIX compliant source */

#define _POSIX_SOURCE 1

/* number of hex columns in table view */

#define NUM_HEX_COLS 4

/* global variables */

int use_cookies = FALSE;
gchar *cookie_id; 

struct RFIDTagType *tagTypes[100];
int tagTypeIdx;

int RFID_PAGE_LENGTH;

int fd;
struct RFIDTag tag;
char readerdevice[255];

static GtkWidget *entryTagID;
static GtkWidget *entryManufacturer;
static GtkWidget *entryTagType;
static GtkWidget *treeview1;
static GtkWidget *statusbar;

static GtkTreeModel *model;

int context_id;

enum
{
    COL_ADR = 0,
    COL_0,
    COL_1,
    COL_2,
    COL_3,
//    COL_4,
//    COL_5,
//    COL_6,
//    COL_7,
    COL_ASCII,
    NUM_COLS
};

int main (int argc, char *argv[])
{
  GtkWidget *windowRFDump;

#ifdef ENABLE_NLS
  bindtextdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  textdomain (GETTEXT_PACKAGE);
#endif

  gtk_set_locale ();
  gtk_init (&argc, &argv);

  add_pixmap_directory (PACKAGE_DATA_DIR "/" PACKAGE "/pixmaps");

  /* code generated by glade */

  windowRFDump = create_windowRFDump();
  gtk_widget_show (windowRFDump);

  /* custom code starts here */

  cookie_id = (gchar*)malloc(32*sizeof(gchar));
  strcpy(cookie_id, "DEADFACE");

  treeview1 = lookup_widget(GTK_WIDGET(windowRFDump), "treeview1");

  entryTagID = lookup_widget(GTK_WIDGET(windowRFDump), "entryTagID");
  entryTagType = lookup_widget(GTK_WIDGET(windowRFDump), "entryTagType");
  entryManufacturer = lookup_widget(GTK_WIDGET(windowRFDump), "entryManufacturer");
  statusbar = lookup_widget(GTK_WIDGET(windowRFDump), "statusbar");

  context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), "RFDump Statusbar");

  g_signal_connect(G_OBJECT(windowRFDump), "delete_event", G_CALLBACK(delete_event), NULL);
  g_timeout_add (1000, timeout_callback, (gpointer)"workThread");

  printf("\n\nRFDump V1.3\n\n");
  printf("Copyright (c) 2005 by DN-Systems GmbH\n");
  printf("The Tool for RFID Supply Chain Auditing and Security\n");
  printf("For more information see www.rfdump.org\n\n");

  /* read about tag types defined in rfd_types.xml */

  int noTagTypes = readTagTypes(tagTypes);

  /* get device property from config file */

  getProperty("DEV", readerdevice);

  if (argc == 2)
  {
    strcpy(readerdevice, argv[1]);
  } else if (argc > 2) {
    printf("Usage: rfdump [tty]\n\n");
    exit(1);
  }

  gtk_statusbar_push(GTK_STATUSBAR(statusbar), context_id, "INITIALIZING READER");

  fd = rfidConnectReader(readerdevice);
  rfidReset(fd);
  rfidInitTag(&tag);

  gtk_statusbar_push(GTK_STATUSBAR(statusbar), context_id, "SCANNING FOR TAGS");

  /* create list view and model */

  GtkListStore *liststore;
  GtkTreeIter   iter;

  liststore = gtk_list_store_new(NUM_COLS, 
        G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING,
        G_TYPE_STRING);//, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);

  GtkTreeViewColumn   *col;
  GtkCellRenderer     *renderer;

  /* --- Column #1 (Adr) --- */

  col = gtk_tree_view_column_new();
  gtk_tree_view_column_set_title(col, "Adr");

  /* pack tree view column into tree view */

  gtk_tree_view_append_column(GTK_TREE_VIEW(treeview1), col);
  renderer = gtk_cell_renderer_text_new();

  /* pack cell renderer into tree view column */

  gtk_tree_view_column_pack_start(col, renderer, TRUE);
  g_object_set_data(G_OBJECT(renderer), "my_column_num", GUINT_TO_POINTER(COL_ADR));
  gtk_tree_view_column_add_attribute(col, renderer, "text", COL_ADR);
  g_signal_connect(renderer, "edited", (GCallback)cell_edited_callback, liststore);

  /* --- Columns #2 - #5 --- (Hex) */

  int colNo = 0;

  char headers[][NUM_HEX_COLS] = {"HEX", "", "", ""};

  for (colNo=1; colNo<=NUM_HEX_COLS; colNo++) {

    col = gtk_tree_view_column_new();

    gtk_tree_view_column_set_title(col, &(headers[colNo-1][0]));

    /* pack tree view column into tree view */

    gtk_tree_view_append_column(GTK_TREE_VIEW(treeview1), col);

    renderer = gtk_cell_renderer_text_new();

    /* pack cell renderer into tree view column */

    gtk_tree_view_column_pack_start(col, renderer, TRUE);

    /* set 'cell-background' property of the cell renderer */

    g_object_set(renderer,
               "cell-background", "Blue",
               "cell-background-set", TRUE,
               NULL);

    g_object_set(renderer, "editable", TRUE, NULL);
    g_object_set_data(G_OBJECT(renderer), "my_column_num", GUINT_TO_POINTER(colNo));
    gtk_tree_view_column_add_attribute(col, renderer, "text", colNo);
    g_signal_connect(renderer, "edited", (GCallback)cell_edited_callback, liststore);
  }

  /* --- Column #6 (ASCII) --- */

  col = gtk_tree_view_column_new();
  gtk_tree_view_column_set_title(col, "ASCII");

  /* pack tree view column into tree view */

  gtk_tree_view_append_column(GTK_TREE_VIEW(treeview1), col);
  renderer = gtk_cell_renderer_text_new();

  /* pack cell renderer into tree view column */

  gtk_tree_view_column_pack_start(col, renderer, TRUE);
  g_object_set(renderer, "editable", TRUE, NULL);
  g_object_set(renderer, "family", "Monospace", NULL);
  g_object_set_data(G_OBJECT(renderer), "my_column_num", GUINT_TO_POINTER(COL_ASCII));
  gtk_tree_view_column_add_attribute(col, renderer, "text", COL_ASCII);
  g_signal_connect(renderer, "edited", (GCallback)cell_edited_callback, liststore);

  model = GTK_TREE_MODEL(liststore);
  gtk_tree_view_set_model(GTK_TREE_VIEW(treeview1), GTK_TREE_MODEL(liststore));

  g_object_unref(liststore);

  /* Now the model will be destroyed when the tree view is destroyed */

  gtk_main();
  return 0;
}

void updateGUI(int adr, gchar *new_text) {
  
  int row_number = (int)adr/NUM_HEX_COLS;
  int column_number = (adr%NUM_HEX_COLS)+1;

  char path_string[255];
  sprintf(path_string, "%d", row_number);

  GtkTreePath *path = gtk_tree_path_new_from_string(path_string);

  GtkTreeIter iter;  
  gtk_tree_model_get_iter(model, &iter, path);
  gtk_list_store_set((GtkListStore*)model, &iter, column_number, new_text, -1);

  /* update ASCII */

  gchar *ascii;
  gtk_tree_model_get(model, &iter, COL_ASCII, &ascii, -1);

  int i;

  for (i=0; i<strlen(new_text)/2; i++)
  {
    int asciiCode = getCharValue(new_text[2*i])*16 + getCharValue(new_text[2*i+1]);

      if ((asciiCode>=(int)'A' && asciiCode<=(int)'Z') ||
          (asciiCode>=(int)'a' && asciiCode<=(int)'z') ||
          (asciiCode>=(int)'0' && asciiCode<=(int)'9')) 
      ascii[((column_number-1)*(strlen(new_text)/2))+i] = (char)asciiCode;
    else 
      ascii[((column_number-1)*(strlen(new_text)/2))+i] = '.';
  }

  gtk_list_store_set((GtkListStore*)model, &iter, COL_ASCII, ascii, -1);
}

void cell_edited_callback (GtkCellRendererText *cell, gchar *path_string, gchar *new_text, 
  gpointer user_data) {

  printf("CELL EDITED:%s\n", new_text);

  /* get column */

  guint column_number = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(cell), "my_column_num"));
  //printf("COLUMN:%d\n", column_number);

  /* do padding / cropping for ASCII field if necessary */

  if (column_number == COL_ASCII) {

    if (strlen(new_text)>NUM_HEX_COLS*RFID_PAGE_LENGTH)
    {
      new_text[NUM_HEX_COLS*RFID_PAGE_LENGTH] = '\0';
    } else if ((strlen(new_text)<NUM_HEX_COLS*RFID_PAGE_LENGTH)) {
      int i;

      gchar *tmp_new_text = (gchar*)malloc((NUM_HEX_COLS*RFID_PAGE_LENGTH)+1);
      strcpy(tmp_new_text, new_text);

      for (i=0; i<(NUM_HEX_COLS*RFID_PAGE_LENGTH)-strlen(new_text); i++)
      {
        strcat(tmp_new_text,".");
      }

      new_text = tmp_new_text;
    }

  /* do padding / cropping for HEX field if necessary */

  } else if (column_number>=COL_0 && column_number<COL_ASCII) {

    if (strlen(new_text)>2*RFID_PAGE_LENGTH)
    {
      new_text[2*RFID_PAGE_LENGTH] = '\0';
    } else if ((strlen(new_text)<2*RFID_PAGE_LENGTH)) {

      int i;

      gchar *tmp_new_text = (gchar*)malloc((2*RFID_PAGE_LENGTH)+1);
      strcpy(tmp_new_text, new_text);

      for (i=0; i<(2*RFID_PAGE_LENGTH)-strlen(new_text); i++)
      {
        strcat(tmp_new_text,"0");
      }

      new_text = tmp_new_text;
    }
  }

  GtkTreePath *path = gtk_tree_path_new_from_string(path_string);

  /* get row */

  gint *indices = gtk_tree_path_get_indices(path);
  int row_number = indices[0];
  //printf("ROW:%d\n", row_number);

  GtkTreeIter iter;  
  gtk_tree_model_get_iter(model, &iter, path);

  /* update model */

  gtk_list_store_set((GtkListStore*)model, &iter, column_number, new_text, -1);

  /* write data to RFID tag */

  /* HEX editor */

  if (column_number>=COL_0 && column_number<COL_ASCII) {

    int intAdr = row_number*NUM_HEX_COLS + column_number-1;    
    char hexAdr[3];

    sprintf(hexAdr,"%02x",intAdr);

    rfidWritePage(fd, hexAdr, new_text);

    /* update ASCII */

    gchar *ascii;
    gtk_tree_model_get(model, &iter, COL_ASCII, &ascii, -1);

    int i;

    for (i=0; i<strlen(new_text)/2; i++)
    {
      int asciiCode = getCharValue(new_text[2*i])*16 + getCharValue(new_text[2*i+1]);

      if ((asciiCode>=(int)'A' && asciiCode<=(int)'Z') ||
          (asciiCode>=(int)'a' && asciiCode<=(int)'z') ||
          (asciiCode>=(int)'0' && asciiCode<=(int)'9')) 
        ascii[((column_number-1)*(strlen(new_text)/2))+i] = (char)asciiCode;
      else 
        ascii[((column_number-1)*(strlen(new_text)/2))+i] = '.';
    }

    gtk_list_store_set((GtkListStore*)model, &iter, COL_ASCII, ascii, -1);

  /* ASCII editor */

  } else if (column_number == COL_ASCII) {

    int i;
    char page[17];
    char pageByte[3];
    char pageAdr[3];
    int hexCol = 1;

    page[0] = '\0';

    for (i=0; i<strlen(new_text); i++) {

      sprintf(pageByte, "%02X", (int)new_text[i]);
      strcat(page, pageByte);

      if ((i+1)%RFID_PAGE_LENGTH == 0) {

        int adr = (row_number*NUM_HEX_COLS)+(int)(i/RFID_PAGE_LENGTH);

        sprintf(pageAdr, "%02x", adr);

        gchar *hexValue = (gchar*)malloc(sizeof(page)+1);
        strcpy(hexValue, page);

        rfidWritePage(fd, pageAdr, page);
        page[0] = '\0';

        /* update HEX */

        //printf("NEW VALUE:%s\n", hexValue);
        
        gtk_list_store_set((GtkListStore*)model, &iter, hexCol, hexValue, -1);
        hexCol++;
      }
    }
  }
}

gboolean delete_event(GtkWidget *widget, GdkEvent *event, gpointer data) {

    gtk_main_quit();
    return FALSE;
}

gint timeout_callback(gpointer data) {

  static char line[255];
  static char tagID[255];
  static char newTagID[255];
  static char page[255];
  static char pageNo[255];

  static int nothing_in_field = FALSE;

  //printf("TIMER WRITING: %s\n", RFID_CMD_SELECT);
  write(fd, RFID_CMD_SELECT, 1);
  rfidReadLine(fd, line);
  printf("TAG: %s\n", line);

  tagTypeIdx = getTagType(line, tagTypes);

  if (strlen(line)>1) {

    rfidGetTagID(line, newTagID);

    if (strcmp(newTagID, tagID)) {

      strcpy(tagID,newTagID);

      strcpy(tag.tagID, tagID);
      strcpy(tag.tagTypeName, tagTypes[tagTypeIdx]->tagTypeName);
      strcpy(tag.tagManufacturerName, tagTypes[tagTypeIdx]->tagManufacturerName);
      tag.memSize = tagTypes[tagTypeIdx]->memSize;
      tag.pageSize = tagTypes[tagTypeIdx]->pageSize;
      RFID_PAGE_LENGTH = tagTypes[tagTypeIdx]->pageSize;

      gtk_entry_set_text((GtkEntry*)entryTagID, tagID);
      gtk_entry_set_text((GtkEntry*)entryTagType, tag.tagTypeName);
      gtk_entry_set_text((GtkEntry*)entryManufacturer, tag.tagManufacturerName);

      /* read memory */

      gtk_statusbar_push(GTK_STATUSBAR(statusbar), context_id, "READING TAG MEMORY");
      printf("READING MEMORY...");
      gtk_list_store_clear((GtkListStore*)model);
      rfidReadAllPages(fd, &tag);
      gtk_statusbar_pop(GTK_STATUSBAR(statusbar), context_id);
      printf("DONE\n");

      char *row[10];
      int i, k, adr = 0;

      for (k=0; k<(tag.memSize/NUM_HEX_COLS)+1; k++) {

        row[0] = (char*)malloc(sizeof("00")+1);
        sprintf(row[0],"%02X",(int)k*NUM_HEX_COLS);

        row[COL_ASCII] = (char*)malloc(255+1);
        *(row[COL_ASCII]) = '\0';

        for (i=0; i<NUM_HEX_COLS; i++) {

          adr = k*NUM_HEX_COLS+i;

          if (tag.mem[adr] != NULL) {
            
            row[i+1] = (char*)malloc(strlen(tag.mem[adr])+1);
            strcpy(row[i+1], tag.mem[adr]);
            strcat(row[COL_ASCII], tag.mem[adr]);

          } else {

            row[i+1] = (char*)malloc(strlen("--------")+1);
            strcpy(row[i+1], "--------");
            strcat(row[COL_ASCII], "........");
          }
        }

        for (i=0; i<strlen(row[COL_ASCII])/2; i++) {

          int asciiCode = getCharValue(row[COL_ASCII][2*i])*16 + getCharValue(row[COL_ASCII][2*i+1]);

          if ((asciiCode>=(int)'A' && asciiCode<=(int)'Z') ||
              (asciiCode>=(int)'a' && asciiCode<=(int)'z') ||
              (asciiCode>=(int)'0' && asciiCode<=(int)'9')) 
            row[COL_ASCII][i] = (char)asciiCode;
          else 
            row[COL_ASCII][i] = '.';
        }

        row[COL_ASCII][i] = '\0';

        GtkTreeIter iter;

        /* append an empty row to the list store, iter will point to the new row */

        gtk_list_store_append((GtkListStore*)model, &iter);

        /* fill fields with data */

        gtk_list_store_set((GtkListStore*)model, &iter, 0, row[0], 1, row[1], 2, row[2], 3, row[3], 4, 
          row[4], 5, row[5], /*6, row[6], 7, row[7], 8, row[8], 9, row[9],*/ -1);
      }
    }

    printf("\n");

    if (nothing_in_field) {

      /* cookie handling */

      nothing_in_field = FALSE;

      if (use_cookies)
      {
        sprintf(pageNo,"%02x", tag.memSize-2);
        rfidReadPage(fd, pageNo, page);

        int i;

        for (i=0; i<strlen(cookie_id); i++)
        {
          cookie_id[i] = toupper(cookie_id[i]);
        }

        if (!strcmp(page, cookie_id)) {

          printf("COOKIE FOUND\n");

          int counter = rfidIncrementCookie(fd, &tag);
          sprintf(page,"%08d",counter);
          updateGUI(tag.memSize-1, page);
          //gtk_entry_set_text(GTK_ENTRY(entryCounter), page);

        } else {

          printf("NO COOKIE FOUND - INITIALIZING\n");

          rfidInitCookie(fd, &tag, cookie_id);
          updateGUI(tag.memSize-2, cookie_id);
          updateGUI(tag.memSize-1, "00000000");
        }
      }
    }

  } else {
    nothing_in_field = TRUE;
  }

  printf("\n");
}

int getCharValue(char c) {
 
  switch (c) {
    case '0' : return 0;
    case '1' : return 1;
    case '2' : return 2;
    case '3' : return 3;
    case '4' : return 4;
    case '5' : return 5;
    case '6' : return 6;
    case '7' : return 7;
    case '8' : return 8;
    case '9' : return 9;
    case 'A' : return 10;
    case 'B' : return 11;
    case 'C' : return 12;
    case 'D' : return 13;
    case 'E' : return 14;
    case 'F' : return 15;
  }
}
